from typing import List, Union, Callable, Tuple

from buils.BuildConfig import BuildConfig
from buils.early.defenceRoaches import DefenceRoach
from sc2.ids.upgrade_id import UpgradeId

from buils.BaseBuild import BaseBuild, Gas, Minerals, BuildOrder, Tech
from buils.build import Build
from buils.early.normal_macro import NormalMacro
from buils.early.turtle import Turtle
from buils.lategame.lurkers import LurkersBuild
from buils.midgame.roachplay import RoachPlay
from buils.util import build_workers, max_workers_reached, skip_new_base, is_not_lategame, is_not_midgame, \
    is_not_earlygame
from sc2 import UnitTypeId, Race

from sharpy.plans.acts import ActBase, PositionBuilding, DefensePosition, ActTech, Expand
from sharpy.plans.acts.zerg import ZergUnit, AutoOverLord, MorphOverseer, MorphLair
from sharpy.plans.build_step import Step
from sharpy.plans.require import (
    RequiredUnitReady,
    EnemyUnitExists, RequiredSupply, RequireCustom, Time, Supply, RequiredGas, UnitReady, All, Time,
)
from sharpy.plans.sequential_list import SequentialList

from sharpy.plans.tactics.zerg import CounterTerranTie
from plans.act_expand import ActExpand
from plans.step_gas import StepBuildGas
import math

class BuildDrones(BuildOrder):
    def __init__(self):
        self.drones = ZergUnit(UnitTypeId.DRONE)

        super().__init__(self.drones)

    async def execute(self) -> bool:
        self.drones.to_count = math.floor(self.knowledge.build_config.to_drone_amount * 100)
        return await super().execute()


class BuildDecisionAct(ActBase):

    def __init__(
            self,
            knowledge: "Knowledge",
            opening: Build,
            midgame: Build,
            lategame: Build,
    ):
        """
        Build order is a list of actions that are executed sequentially, but they are not blocking.
        When every act in build order returns true, so will also build order.

        """
        super().__init__()

        self.checked_for_cheese = False

        self.opening: Build = opening if opening is not None else NormalMacro()
        self.midgame: Build = midgame if midgame is not None else RoachPlay()
        self.lategame: Build = lategame if lategame is not None else LurkersBuild()

        self.createBuild()

    async def debug_draw(self):
        for order in self.orders:
            await order.debug_draw()

    def createBuild(self):

        self.orders = [
            CounterTerranTie([
                SequentialList(
                    # Overlords
                    Step(Supply(13), ZergUnit(UnitTypeId.OVERLORD, 2, priority=True)),
                    Step(Supply(20), ZergUnit(UnitTypeId.OVERLORD, 3, priority=True)),
                    AutoOverLord(),
                ),
                self.opening.build(),

                SequentialList(
                    # Workers
                    Step(build_workers, ZergUnit(UnitTypeId.DRONE, 19)),
                ),

                SequentialList(
                    Step(UnitReady(UnitTypeId.DRONE, 19), None),
                    Step(RequireCustom(lambda k: self.opening.is_done()), None),
                    self.midgame.build(),
                    Step(UnitReady(UnitTypeId.LAIR), MorphOverseer(2)),
                ),

                SequentialList(
                    Step(RequireCustom(lambda k: self.opening.is_done()), None),
                    Step(
                        All([UnitReady(UnitTypeId.HATCHERY), Gas(150), Minerals(100)]),
                        Tech(UpgradeId.OVERLORDSPEED),
                    ),
                ),

                SequentialList(
                    Step(
                        UnitReady(UnitTypeId.SPAWNINGPOOL),
                        PositionBuilding(UnitTypeId.SPINECRAWLER, DefensePosition.FarEntrance, to_count=1,
                                         to_base_index=0),
                        skip=Time(60*2),
                        skip_until=EnemyUnitExists(UnitTypeId.ZERGLING, count=5)
                    ),
                    Step(
                        UnitReady(UnitTypeId.SPAWNINGPOOL),
                        PositionBuilding(UnitTypeId.SPINECRAWLER, DefensePosition.FarEntrance, to_count=2,
                                         to_base_index=0),
                        skip=Time(60*2),
                        skip_until=EnemyUnitExists(UnitTypeId.ZERGLING, count=10)
                    ),
                ),

                SequentialList(
                    Step(RequireCustom(lambda k: k.enemy_units_manager.enemy_cloak_trigger), MorphLair()),
                    Step(UnitReady(UnitTypeId.LAIR), MorphOverseer(2)),
                ),

                SequentialList(
                    # Some standard air defence
                    Step(RequireCustom(lambda k: self.opening.is_done()), None),
                    Step(RequireCustom(lambda k: k.enemy_race != "Zerg" and self.get_count(UnitTypeId.DRONE) > 40),
                         PositionBuilding(UnitTypeId.SPORECRAWLER, DefensePosition.CenterMineralLine, to_base_index=0)),
                    Step(RequireCustom(lambda k: k.enemy_race != "Zerg" and self.get_count(UnitTypeId.DRONE) > 39),
                         PositionBuilding(UnitTypeId.SPORECRAWLER, DefensePosition.CenterMineralLine, to_base_index=1,
                                          to_count=2)),
                    Step(RequireCustom(lambda k: k.enemy_race != "Zerg" and self.get_count(UnitTypeId.DRONE) > 38),
                         PositionBuilding(UnitTypeId.SPORECRAWLER, DefensePosition.CenterMineralLine, to_base_index=2,
                                          to_count=3)),
                ),
                SequentialList(
                    Step(UnitReady(UnitTypeId.DRONE, 80), None),
                    Step(RequireCustom(lambda k: self.midgame.is_done()), None),
                    self.lategame.build(),
                ),

                self.opening.tech(),
                SequentialList(
                    Step(UnitReady(UnitTypeId.DRONE, 19), None),
                    Step(RequireCustom(lambda k: self.opening.is_done()), None),
                    self.midgame.tech(),
                ),
                SequentialList(
                    Step(UnitReady(UnitTypeId.DRONE, 80), None),
                    Step(RequireCustom(lambda k: self.midgame.is_done()), None),
                    self.lategame.tech(),
                ),

                SequentialList(
                    Step(RequireCustom(lambda k: self.opening.is_done()), None),
                    Step(None, ZergUnit(UnitTypeId.QUEEN, 10), skip=RequireCustom(
                        lambda k: len(k.unit_cache.own({UnitTypeId.HATCHERY, UnitTypeId.LAIR, UnitTypeId.HIVE})) < len(
                            k.unit_cache.own({UnitTypeId.QUEEN})) and \
                                  # or build queens if enemy is on few bases and you have minerals to spare but not any gas
                                  (len(k.enemy_townhalls) \
                                    > 2 or (k.available_mineral < 300 or k.available_gas > 50))
                    )),
                    Step(build_workers, ZergUnit(UnitTypeId.DRONE, 80), skip=RequireCustom(
                        lambda k: max_workers_reached(k) or (k.available_gas < 25 and k.available_mineral > 50))),
                    # more bases and gas after 40 drones
                    Step(All([Time(5 * 60), UnitReady(UnitTypeId.DRONE, 40)]), None),
                    Step(build_workers, Expand(10), skip=skip_new_base),
                    # More gas geysers
                    StepBuildGas(20, \
                      skip=RequireCustom(lambda k: len(k.unit_cache.own({UnitTypeId.EXTRACTOR})) * 12 > len(k.unit_cache.own({UnitTypeId.DRONE})) < 32 or \
                       (k.available_gas > 300 and k.available_mineral < 600) or \
                        len(k.unit_cache.own({UnitTypeId.HATCHERY, UnitTypeId.LAIR, UnitTypeId.HIVE}).ready) * 1.5 < \
                           len(k.unit_cache.own({UnitTypeId.EXTRACTOR, UnitTypeId.EXTRACTORRICH})))),
                ),

                SequentialList(
                    Step(RequireCustom(lambda k: self.opening.is_done()), None),
                    # Workers
                    Step(build_workers, BuildDrones(), skip=max_workers_reached),
                    #Step(None, BuildDrones(), skip=max_workers_reached),
                ),

                Step(RequireCustom(lambda k: self.midgame.is_done()), self.lategame.units(), skip=is_not_lategame),
                Step(RequireCustom(lambda k: self.opening.is_done()), self.midgame.units(), skip=is_not_midgame),
                Step(None, self.opening.units(), skip=is_not_earlygame),
            ]),
        ]

    async def change_build(self, opening: Build = None, midgame: Build = None, lategame: Build = None):
        if opening is not None:
            self.opening = opening
        if midgame is not None:
            self.midgame = midgame
        if lategame is not None:
            self.lategame = lategame

        self.createBuild()

        for order in self.orders:
            await order.start(self.knowledge)

    async def start(self, knowledge: "Knowledge"):
        await super().start(knowledge)

        for order in self.orders:
            await order.start(knowledge)

    async def execute(self) -> bool:

        if not isinstance(self.opening, Turtle) and self.knowledge.build_detector.rush_detected:
            await self.change_build(opening=Turtle())

        # If enemy zerg rushes gas or spawningpool
        # Or against enemy proxy buildings
        if not isinstance(self.opening, DefenceRoach) and self.knowledge.ai.time < 130 and \
                (
                  (
                    self.knowledge.unit_cache.enemy(UnitTypeId.EXTRACTOR).ready.amount > 0 or
                    (self.knowledge.unit_cache.enemy(UnitTypeId.SPAWNINGPOOL).exists and (1 - self.knowledge.unit_cache.enemy(UnitTypeId.SPAWNINGPOOL).first.build_progress)*46 + self.knowledge.ai.time < 90)
                  ) or
                  (
                    self.knowledge.unit_cache.enemy({UnitTypeId.GATEWAY, UnitTypeId.BARRACKS}).exists and
                    self.knowledge.unit_cache.enemy({UnitTypeId.GATEWAY, UnitTypeId.BARRACKS}).closest_distance_to(self.knowledge.own_main_zone.center_location) < self.knowledge.rush_distance / 2
                  )
                ):
            await self.change_build(opening=DefenceRoach(self.knowledge))

        # If enemy makes a huge army early
        if not isinstance(self.opening, DefenceRoach) and self.knowledge.ai.time < 3*60+30 and \
                (
                    self.knowledge.unit_cache.enemy(UnitTypeId.ZERGLING).amount > 9 or
                    self.knowledge.unit_cache.enemy(UnitTypeId.ZEALOT).amount > 9 or
                    self.knowledge.unit_cache.enemy(UnitTypeId.MARINE).amount > 9
                ):
            await self.change_build(opening=DefenceRoach(self.knowledge))

        result = True
        for order in self.orders:
            if not await order.execute():
                result = False

        return result
