from typing import Union, Callable, List

from sc2 import UnitTypeId, Race
from sc2.ids.upgrade_id import UpgradeId
from sc2.unit import Unit
from sharpy.knowledges import KnowledgeBot, Knowledge
from sharpy.plans.zerg import *


class RoachAndRavager(BuildOrder):
    def __init__(self):
        self.ravager = ZergUnit(UnitTypeId.RAVAGER)
        self.roach = ZergUnit(UnitTypeId.ROACH)
        super().__init__(self.ravager, self.roach)

    async def execute(self) -> bool:
        return await super().execute()


class RoachRavagerBuild(BuildOrder):
    def __init__(self):
        super().__init__(
            SequentialList(
                # Overlords
                Step(RequiredSupply(13), ZergUnit(UnitTypeId.OVERLORD, 2, priority=True)),
                Step(RequiredSupply(20), ZergUnit(UnitTypeId.OVERLORD, 3, priority=True)),
                AutoOverLord(),
            ),
            SequentialList(
                # Opener
                Step(RequiredSupply(17), ActExpand(2)),
                Step(
                    RequiredSupply(17),
                    PositionBuilding(UnitTypeId.SPAWNINGPOOL, DefensePosition.BehindMineralLineLeft, 0),
                ),
                StepBuildGas(1, RequiredSupply(22)),
                ActUnit(UnitTypeId.ZERGLING, UnitTypeId.LARVA, 4),
                ZergUnit(UnitTypeId.QUEEN, 2),
                Step(self.build_workers, None, skip=Time(8 * 60)),
                Step(
                    UnitExists(UnitTypeId.DRONE, 32, include_pending=True),
                    StepBuildGas(4),
                ),
                Step(RequiredUnitReady(UnitTypeId.LAIR), MorphOverseer(2)),
                Step(RequiredMinerals(600), ActExpand(3)),
                StepBuildGas(6),
            ),
            SequentialList(
                # Workers
                #ZergUnit(UnitTypeId.DRONE, 22),
                Step(self.build_workers, ZergUnit(UnitTypeId.DRONE, 50), skip=self.max_workers_reached),
            ),
            SequentialList(
                Step(self.build_workers, None, skip=Time(8 * 60)),
                Step(RequireCustom(lambda k: k.enemy_units_manager.enemy_cloak_trigger), MorphLair()),
                Step(RequiredUnitReady(UnitTypeId.LAIR), MorphOverseer(2)),
            ),
            SequentialList(
                # Tech
                Step(RequiredUnitReady(UnitTypeId.SPAWNINGPOOL), ActBuilding(UnitTypeId.ROACHWARREN)),
                Step(self.build_workers, None, skip=Time(8 * 60)),
                Step(RequiredUnitReady(UnitTypeId.HATCHERY), MorphLair(), skip=RequiredUnitReady(UnitTypeId.HIVE)),
                Step(RequiredAll([RequiredUnitReady(UnitTypeId.ROACHWARREN), RequiredUnitReady(UnitTypeId.LAIR)]),
                     ActTech(UpgradeId.GLIALRECONSTITUTION)),
            ),
            SequentialList(
            # Some standard air defence
                Step(RequireCustom(lambda k: k.enemy_race != "zerg" and self.get_count(UnitTypeId.DRONE) > 40), PositionBuilding(UnitTypeId.SPORECRAWLER, DefensePosition.CenterMineralLine, to_base_index=0)),
                Step(RequireCustom(lambda k: k.enemy_race != "zerg" and self.get_count(UnitTypeId.DRONE) > 39), PositionBuilding(UnitTypeId.SPORECRAWLER, DefensePosition.CenterMineralLine, to_base_index=1, to_count=2)),
                Step(RequireCustom(lambda k: k.enemy_race != "zerg" and self.get_count(UnitTypeId.DRONE) > 38), PositionBuilding(UnitTypeId.SPORECRAWLER, DefensePosition.CenterMineralLine, to_base_index=2, to_count=3)),
            ),
            SequentialList(
                # more bases, because too little gas
                Step(Time(7 * 60), None, skip=RequireCustom(lambda k: k.available_gas < 25)),
                Step(None, ActExpand(10), skip=self.skip_new_base),
                StepBuildGas(20, skip=RequireCustom(lambda k: len(self.cache.own({UnitTypeId.DRONE})) < 32 or len(self.cache.own({UnitTypeId.HATCHERY, UnitTypeId.LAIR, UnitTypeId.HIVE}).ready) * 2 < len(self.cache.own(
                    {UnitTypeId.EXTRACTOR, UnitTypeId.EXTRACTORRICH})))),
                Step(None, ZergUnit(UnitTypeId.QUEEN, 10), RequireCustom(lambda k: len(self.cache.own({UnitTypeId.HATCHERY, UnitTypeId.LAIR, UnitTypeId.HIVE})) < len(self.cache.own({UnitTypeId.QUEEN})))),
                Step(self.build_workers, ZergUnit(UnitTypeId.DRONE, 80), skip=RequireCustom(lambda k: self.max_workers_reached(k) or (k.available_gas < 25 and k.available_mineral > 50))),
            ),
            SequentialList(
                # Upgrades after a while
                Step(self.build_workers, None, skip=Time(8 * 60)),
                Step(Time(60 * 6), ActBuilding(UnitTypeId.EVOLUTIONCHAMBER, to_count=2)),
                Step(RequiredUnitReady(UnitTypeId.EVOLUTIONCHAMBER), ActTech(UpgradeId.ZERGMISSILEWEAPONSLEVEL1)),
                Step(RequiredUnitReady(UnitTypeId.EVOLUTIONCHAMBER), ActTech(UpgradeId.ZERGGROUNDARMORSLEVEL1)),
                Step(RequiredUnitReady(UnitTypeId.LAIR), ActTech(UpgradeId.ZERGMISSILEWEAPONSLEVEL2)),
                Step(RequiredUnitReady(UnitTypeId.LAIR), ActTech(UpgradeId.ZERGGROUNDARMORSLEVEL2)),
                Step(RequiredUnitReady(UnitTypeId.HIVE), ActTech(UpgradeId.ZERGMISSILEWEAPONSLEVEL3)),
                Step(RequiredUnitReady(UnitTypeId.HIVE), ActTech(UpgradeId.ZERGGROUNDARMORSLEVEL3)),
            ),
            Step(
                None,
                SequentialList(
                    # Units
                    Step(None, RoachAndRavager(), skip_until=RequiredUnitReady(UnitTypeId.ROACHWARREN)),
                    ZergUnit(UnitTypeId.ZERGLING),
                ),
            ),
        )

    def build_workers(self, knowledge: Knowledge) -> bool:

        for zone in knowledge.zone_manager.expansion_zones:
            if zone.is_ours and zone.is_under_attack:
                return False

        return knowledge.game_analyzer.army_can_survive

    def max_workers_reached(self, knowledge: Knowledge) -> bool:
        count = 1
        for townhall in self.ai.townhalls:  # type: Unit
            if townhall.is_ready:
                count += townhall.ideal_harvesters
            elif townhall.is_ready > 0.6:
                count += 16
        for gas in self.ai.gas_buildings:  # type: Unit
            count += gas.ideal_harvesters

        if len(knowledge.unit_cache.own(UnitTypeId.DRONE)) >= count:
            return True

        return False

    def skip_new_base(self, knowledge: Knowledge) -> bool:
        count = 1
        for townhall in self.ai.townhalls:  # type: Unit
            if townhall.is_ready:
                count += townhall.ideal_harvesters
            else:
                count += 16
        for gas in self.ai.gas_buildings:  # type: Unit
            count += gas.ideal_harvesters

        # count += 16 * self.ai.already_pending(UnitTypeId.HATCHERY)
        if 0 < self.ai.already_pending(UnitTypeId.HATCHERY) < 0.5:
            return True

        if (len(knowledge.unit_cache.own(UnitTypeId.DRONE)) > count or knowledge.available_mineral > 500) and len(knowledge.unit_cache.own(UnitTypeId.DRONE)) > 28:
            return not knowledge.game_analyzer.army_can_survive

        return True

