import math
from pathlib import Path

from buils.BuildConfig import BuildConfig
from buils.DefinedBuild import BUILDS_ENABLED
from config import get_version
from randomizer import WeightedRandomizer
from sharpy.managers import ManagerBase
from sc2 import UnitTypeId, Dict, Result
import os
import jsonpickle

class BuildDataManager(ManagerBase):

    def __init__(
            self,
    ):
        """
        Build order is a list of actions that are executed sequentially, but they are not blocking.
        When every act in build order returns true, so will also build order.

        """
        super().__init__()

        self.file_name = "data" + os.sep + "my_data" + ".json"
        self.game_data = {}
        self.data = None


    async def start(self, knowledge: "Knowledge"):
        await super().start(knowledge)
        self.knowledge = knowledge

        self.game_data['my_race'] = knowledge.my_race
        self.game_data['version'] = get_version()
        self.game_data['enemy_race'] = knowledge.enemy_race
        self.game_data['enemy_id'] = knowledge.ai.opponent_id
        self.game_data['data_points'] = []

        self.knowledge.build_config = BuildConfig()

        await self.read_old_data()

    async def read_old_data(self):
        try:
            with open(self.file_name, "r") as handle:
                data = jsonpickle.decode(handle.read())
                self.knowledge.print("data read")

                self.data = []
                if 'data' in data:
                    for match in data['data']:
                        if match['enemy_race'] != self.knowledge.enemy_race:
                            continue
                        self.data.append(match)
                self.knowledge.print("data stored")


        except Exception as e:
            # Don't write if we can't read the current data
            self.print(f"Data read failed: {e}")

    def calculate_data(self):
        own_types_left: Dict[str, int] = {}
        enemy_types_left: Dict[str, int] = {}

        for unit_type, units in self.cache.own_unit_cache.items():  # type: (UnitTypeId, Units)
            type_id = self.unit_values.real_type(unit_type)
            val = own_types_left.get(type_id, 0)
            own_types_left[str(type_id)] = val + units.amount

        for unit_count in self.knowledge.enemy_units_manager.enemy_composition:  # type: UnitCount
            unit_type = unit_count.enemy_type
            val = enemy_types_left.get(unit_type, 0)
            enemy_types_left[str(unit_type)] = val + unit_count.count

        data = {
            'iteration': self.knowledge.iteration,
            'enemy_units': enemy_types_left,
            'my_units': own_types_left,
            'build': self.knowledge.build,
            'build_config': self.knowledge.build_config,
        }
        return data

    async def update(self) -> bool:

        if self.knowledge.iteration % 250 == 0:# and self.knowledge.iteration != 0:
            data = self.calculate_data()
            self.game_data['data_points'].append(data)

            await self.find_best_build()

        return True

    async def post_update(self):
        pass

    async def on_end(self, game_result):
        self.game_data['result'] = game_result
        self.save_data()

    def save_data(self):
        my_file = Path(self.file_name)

        if my_file.is_file():
            old_data = {}
            self.print('yep')
            try:
                with open(self.file_name, "r") as handle:
                    old_data = jsonpickle.decode(handle.read())

                old_data['data'].append(self.game_data)
                try:
                    frozen = jsonpickle.encode(old_data)
                    with open(self.file_name, "w") as handle:
                        handle.write(frozen)
                    self.print(f"Data write done")
                    return
                except Exception as e:
                    self.print(f"Data write failed: {e}")
            except Exception as e:
                # Don't write if we can't read the current data
                self.print(f"Data read failed on save: {e}")
                return

        elif not os.path.exists("data"):
            os.makedirs("data")

        real_data = {}
        real_data['data'] = []
        real_data['data'].append(self.game_data)

        frozen = jsonpickle.encode(real_data)
        try:
            with open(self.file_name, "w") as handle:
                handle.write(frozen)
        except Exception as e:
            self.print(f"Data write failed: {e}")
        self.print(f"Data write done")

    @staticmethod
    def distance_between_data_points(data_point1, data_point2):
        distance_between_data = 0

        if not data_point1 or 'enemy_units' not in data_point1 \
                or not data_point2 or 'enemy_units' not in data_point2 \
                or 'my_units' not in data_point1 \
                or 'my_units' not in data_point2:
            return 9999999

        for id, amount in data_point1['enemy_units'].items():
            if str(id) not in data_point2['enemy_units']:
                distance_between_data += 1
            else:
                distance_between_data += 1 - (
                            min(amount, data_point2['enemy_units'][id]) / max(amount, data_point2['enemy_units'][id]))
        for id, amount in data_point2['enemy_units'].items():
            if id not in data_point1['enemy_units']:
                distance_between_data += 1

        for id, amount in data_point1['my_units'].items():
            if str(id) not in data_point2['my_units']:
                distance_between_data += 1
            else:
                distance_between_data += 1 - (
                            min(amount, data_point2['my_units'][id]) / max(amount, data_point2['my_units'][id]))
        for id, amount in data_point2['my_units'].items():
            if id not in data_point1['my_units']:
                distance_between_data += 1

        return distance_between_data

    @staticmethod
    def chose_semi_random_from_stats(stats_dict, default_chance=0.2, power=3):
        # See if the matches that won changed build soon after this
        # Use that data, what strategy that won after a little while
        chance = {}
        for key, value in stats_dict.items():
            if value[1] + value[0] == 0:
                chance[key] = default_chance
            else:
                # Use pow to make the better builds exponentially better than the less good ones
                chance[key] = pow(0.001 + value[0] / (value[1] + value[0]), power)

        return WeightedRandomizer(chance).random()

    async def find_best_build(self):
        self.knowledge.print("Finding best build")
        close_matches = []

        for match in self.data:
            same_opponent = match['enemy_id'] == self.knowledge.ai.opponent_id and match[
                'enemy_race'] == self.knowledge.enemy_race

            if match['enemy_race'] != self.knowledge.enemy_race:
                continue

            for n in range(len(match['data_points']) - 1):

                data_point = match['data_points'][n]

                # if iteration == iteration (possibly next or last as well
                if 'iteration' not in data_point or abs(self.knowledge.iteration - data_point['iteration']) > 250:
                    continue

                if 'build_config' in data_point:
                    data_point['build_config'] = BuildConfig()

                distance_between_data = 0
                for m in range(len(self.game_data['data_points'][:-3])):
                    my_data = self.game_data['data_points'][-(m+1)]
                    if len(match['data_points']) > n-m > 0:
                        distance_between_data += self.distance_between_data_points(my_data, match['data_points'][n-m]) * (1.0 - 0.2 * m)
                distance_between_data /= 3
                # if distance_between_data > 5:
                #    continue

                next_data_point = match['data_points'][n+1]

                data_point['result'] = match['result']
                next_data_point['result'] = match['result']
                data_point['same_opponent'] = same_opponent
                next_data_point['same_opponent'] = same_opponent
                data_point['distance_between_data'] = distance_between_data
                next_data_point['distance_between_data'] = distance_between_data
                data_point['next'] = next_data_point
                close_matches.append(data_point)

        if len(self.data) == 0:
            self.knowledge.print("no data found")

        self.knowledge.print("data compared, got " + str(len(close_matches)) + " matches")

        if len(close_matches) == 0:
            self.knowledge.print("Got no close matches")

        close_matches = sorted(close_matches, key=lambda k: k['distance_between_data'], reverse=True)[:100]

        # compare with current data / last datap_oint
        # Find statistics of different builds from this point
        builds_stats = {}

        for build in BUILDS_ENABLED:
            builds_stats[build] = [0, 0]

        for n in range(len(close_matches)):
            # Use the follow up as a base for statistics
            data_point = close_matches[n]['next']
            if not data_point['build'] in builds_stats or 'result' not in data_point:
                # self.knowledge.print("Build or result missing" + ', '.join([str(p) + " " + str(v) for p, v in data_point.items()]))
                continue

            if data_point['result'].name == "Victory":
                builds_stats[data_point['build']][0] += 1 * ((len(close_matches) - n) / len(close_matches))
            elif data_point['result'].name == "Defeat":
                builds_stats[data_point['build']][1] += 1 * ((len(close_matches) - n) / len(close_matches))
            else:
                self.knowledge.print("Result is wrong")

        self.knowledge.print("amount of builds: " + ', '.join([p + " " + str(v) for p, v in builds_stats.items()]))
        build = BuildDataManager.chose_semi_random_from_stats(builds_stats)
        # build = "spire-build"

        amount_of_workers = {}
        for n in range(10):
            amount_of_workers[n] = [0, 0]
        for data_point in close_matches:
            # Use the follow up as a base for statistics
            data_point = data_point['next']

            if data_point['build'] == build:
                # self.knowledge.print("build is right - " + ', '.join([str(p) + " " + str(v) for p, v in data_point.items()]))
                if 'build_config' in data_point:
                    # self.knowledge.print("has build config")
                    # group by each ten
                    if data_point['result'].name == "Victory":
                        amount_of_workers[math.floor(10*data_point['build_config'].to_drone_amount)][0] += 1
                    if data_point['result'].name == "Defeat":
                        amount_of_workers[math.floor(10*data_point['build_config'].to_drone_amount)][1] += 1

        self.knowledge.print("amount of workers: " + ', '.join([str(p) + " " + str(v) for p, v in amount_of_workers.items()]))
        worker_amount = BuildDataManager.chose_semi_random_from_stats(amount_of_workers) / 10

        # Change to that strategy
        if not build:
            return
        self.knowledge.print("Chose build: " + build)

        if build not in BUILDS_ENABLED:
            return

        drones = worker_amount

        self.knowledge.print("Changed to build: " + build)
        self.knowledge.print("Going to : " + str(drones) + " drones")
        self.knowledge.build = build
        self.knowledge.build_config = BuildConfig()
        self.knowledge.build_config.to_drone_amount = drones

